#include "string_list.h"
#include <UnitTest++/UnitTest++.h>

using namespace string_list;

TEST(Is_empty_recognizes_empty_vs_non)
{
    List mt;
    CHECK(is_empty(mt));

    List lst = cons("a", mt);
    CHECK(!is_empty(lst));
}

TEST(Is_cons_recognizes_empty_vs_non)
{
    List mt;
    CHECK(!is_cons(mt));

    List lst = cons("a", mt);
    CHECK(is_cons(lst));
}

TEST(Cons_builds_list)
{
    List lst = cons("a", cons("b", nullptr));
    CHECK(is_cons(lst));
    CHECK_EQUAL("a", lst->first);
    CHECK(is_cons(lst->rest));
    CHECK_EQUAL("b", lst->rest->first);
    CHECK(is_empty(lst->rest->rest));
}

TEST(Length_of_0_1_2_3_works)
{
    List lst;
    CHECK_EQUAL(0, length(lst));
    lst = cons("c", lst);
    CHECK_EQUAL(1, length(lst));
    lst = cons("b", lst);
    CHECK_EQUAL(2, length(lst));
    lst = cons("a", lst);
    CHECK_EQUAL(3, length(lst));
}

TEST(Append_empty_empty_yields_empty)
{
    List mt;
    List result = append(mt, mt);
    CHECK(is_empty(result));
}

TEST(Append_empty_non_empty_yields_same)
{
    List mt;
    List lst1 = cons("a", cons("b", mt));
    List lst2 = append(mt, lst1);

    CHECK(is_cons(lst2));
    CHECK_EQUAL("a", lst2->first);
    CHECK(is_cons(lst2->rest));
    CHECK_EQUAL("b", lst2->rest->first);
    CHECK(is_empty(lst2->rest->rest));

    CHECK_EQUAL(lst1, lst2);
}

TEST(Append_non_empty_empty_yields_copy)
{
    List mt;
    List lst1 = cons("a", cons("b", mt));
    List lst2 = append(lst1, mt);

    CHECK(is_cons(lst2));
    CHECK_EQUAL("a", lst2->first);
    CHECK(is_cons(lst2->rest));
    CHECK_EQUAL("b", lst2->rest->first);
    CHECK(is_empty(lst2->rest->rest));

    CHECK(lst1 != lst2);
}

TEST(Append_non_empty_non_empty_yields_half_copy)
{
    List lst1 = cons("a", cons("b", nullptr));
    List lst2 = cons("c", cons("d", nullptr));
    List lst3 = append(lst1, lst2);

    CHECK(lst1 != lst3);

    CHECK(is_cons(lst3));
    CHECK_EQUAL("a", lst3->first);
    CHECK(is_cons(lst3->rest));
    CHECK_EQUAL("b", lst3->rest->first);

    CHECK_EQUAL(lst2, lst3->rest->rest);
}

TEST(Concat_empty_empty_yields_empty)
{
    List lst1;
    List lst2;
    concat(lst1, lst2);
    CHECK(is_empty(lst1));
    CHECK(is_empty(lst2));
}

TEST(Concat_empty_non_empty_yields_alias)
{
    List lst1;
    List lst2 = cons("a", cons("b", nullptr));
    concat(lst1, lst2);

    CHECK_EQUAL(lst1, lst2);
    CHECK(is_cons(lst1));
    CHECK(is_cons(lst1->rest));
    CHECK(is_empty(lst1->rest->rest));
}

TEST(Concat_non_empty_empty_no_change)
{
    List lst1 = cons("a", cons("b", nullptr));
    List lst2;
    concat(lst1, lst2);

    CHECK_EQUAL(2, length(lst1));
    CHECK_EQUAL(0, length(lst2));
}

TEST(Concat_non_empty_non_empty_yields_alias)
{
    List lst1 = cons("a", cons("b", nullptr));
    List lst2 = cons("c", cons("d", nullptr));

    concat(lst1, lst2);

    CHECK_EQUAL(4, length(lst1));
    CHECK_EQUAL(2, length(lst2));

    CHECK(is_cons(lst1->rest));
    CHECK(is_cons(lst1->rest->rest));
    CHECK_EQUAL(lst1->rest->rest, lst2);
}
